/* $Id: main.c,v 1.4 2004/04/21 04:23:43 pohlt Exp $ */

/*############################################################################*/

#include "main.h"
#include "lbm.h"
#include <stdio.h>
#include <stdlib.h>

#include <sys/times.h>
#include <unistd.h>
#include <stdint.h>

#include <sys/stat.h>
#include <fcntl.h>

#include "platform.h"
#include "xil_printf.h"
#include "xlbm_acc.h"
#include "xtime_l.h"
#include "xil_io.h"
#include "xil_cache.h"
/*############################################################################*/
#define printf xil_printf

static LBM_GridPtr srcGrid, dstGrid;
static double* src_bus;
static double* dst_bus;

char str[200];

#define ARM_PLL_CTRL 0x100
#define PLL_BYPASS_FORCE 0x10
#define PLL_PWRDWN 0x2

#define ARM_CLK_CTRL 0x120
#define DIVISOR 0x3F00

#define A9_CPU_RST_CTRL 0x244
#define A9_CLKSTOP0 0x10
#define A9_CLKSTOP1 0x20

#define FPGA0_CLK_CTRL 0x170
#define DIVISOR0 0x03F00
//#define DIVISOR1 0x3F000
#define APER_CLK_CTRL 0x12C
#define UART0_CPU_1XCLKACT 20

//#define USE_BRAM
#define USE_ACC
#define DISABLE_CORE
#define CPU_SLOW

#if defined(USE_ACC)
	static const char config[] = "ACC";
#elif defined(USE_BRAM)
	static const char config[] = "BRAM";
#else
	static const char config[] = "ARM";
#endif

#define NSTEPS 30000
static u32 arm_div = 0x3f << 8;
static u32 fpga0_div0 = 0x3f << 8;
//static u32 fpga0_div1 = 0x3F << 12;

static u32 base = XPAR_PS7_SLCR_0_S_AXI_BASEADDR;
static u32 orig_arm_div = 0x200;
static u32 orig_fpga0_div0 = 0x05 << 8;
//static u32 orig_fpga0_div1 = 0x01 << 12;

/*############################################################################*/

int main() {
	init_platform();Xil_DCacheDisable();
#ifdef DISABLE_CORE
	disable_core();
#endif
	MAIN_Param param;
	int check_results = 0;
	sprintf(str, "Started %s %d arm_freq=%.2f MHz fpga0_freq=%.2f MHz !\r\n",
			config, NSTEPS, 1333.33/(arm_div >> 8), 1000.0/(fpga0_div0 >> 8));
	printf(str);

	param.nTimeSteps     = NSTEPS;
	param.resultFilename = "reference.dat";
	param.action         = (MAIN_Action) 0;
	param.simType        = (MAIN_SimType) 0;
	param.obstacleFilename = "100_100_130_ldc_reduced.of";

	MAIN_printInfo( &param );
#ifdef USE_BRAM
	/* Store grids in BRAM */
	src_bus = (double*)XPAR_BRAM_1_BASEADDR;
	src_bus += MARGIN;
	dst_bus = src_bus + GRID_SIZE;
#else
	/* Store grids in main memory */
	src_bus = (double*)0x1e000000;
	src_bus += MARGIN;
	dst_bus = src_bus + GRID_SIZE;
#endif
	srcGrid = &src_bus;
	dstGrid = &dst_bus;
	MAIN_initialize( &param );


	XLbm_acc_Config config;
	XLbm_acc acc;
	config.Axi4lites_BaseAddress = XPAR_LBM_ACC_TOP_0_S_AXI_AXI4LITES_BASEADDR;

	XLbm_acc_Initialize(&acc, &config);

#ifdef CPU_SLOW
	slowdown_cpu();
#endif
	printf("Start computation! base=%x src=%x dst=%x\r\n", config.Axi4lites_BaseAddress, *srcGrid, *dstGrid);
#ifdef USE_ACC
	u32 srcAddr = ((u32)*srcGrid);
	u32 dstAddr = ((u32)*dstGrid);
	XLbm_acc_SetP_srcgrid(&acc, srcAddr/sizeof(double));
	XLbm_acc_SetP_dstgrid(&acc, dstAddr/sizeof(double));
	XLbm_acc_SetNsteps(&acc, param.nTimeSteps);
	XLbm_acc_Start(&acc);
	while (!XLbm_acc_IsDone(&acc)) {
//		printf("Waiting...\r\n");
	}
#else
	int t;
	for( t = 1; t <= param.nTimeSteps; t++ ) {
		performStreamCollide(*srcGrid, *dstGrid);
		LBM_swapGrids( &srcGrid, &dstGrid );
	}
#endif
	printf("Finish computation! \r\n");
#ifdef CPU_SLOW
	restore_cpu();
#endif

#ifdef USE_ACC
	if (param.nTimeSteps % 2 == 1) {
		printf("time steps: %x\r\n", param.nTimeSteps);
		LBM_swapGrids( &srcGrid, &dstGrid );
	}
#endif
	MAIN_finalize( &param );

	printf("Simulation Complete\r\n");

	cleanup_platform();
	return check_results;
}

/*############################################################################*/

void MAIN_initialize( const MAIN_Param* param ) {
	LBM_initializeGrid( *srcGrid );
	LBM_initializeGrid( *dstGrid );

	if( param->obstacleFilename != NULL ) {
		LBM_loadObstacleFile( *srcGrid, param->obstacleFilename );
		LBM_loadObstacleFile( *dstGrid, param->obstacleFilename );
	}

	if( param->simType == CHANNEL ) {
		LBM_initializeSpecialCellsForChannel( *srcGrid );
		LBM_initializeSpecialCellsForChannel( *dstGrid );
	}
	else {
		LBM_initializeSpecialCellsForLDC( *srcGrid );
		LBM_initializeSpecialCellsForLDC( *dstGrid );
	}

	LBM_showGridStatistics( *srcGrid );
}

/*############################################################################*/

void MAIN_finalize( const MAIN_Param* param ) {
	LBM_showGridStatistics( *srcGrid );

#ifndef USE_BRAM
	LBM_freeGrid(srcGrid);
	LBM_freeGrid(dstGrid);
#endif
}

/*############################################################################*/
void MAIN_printInfo( const MAIN_Param* param ) {
	const char actionString[3][32] = {"nothing", "compare", "store"};
	const char simTypeString[3][32] = {"lid-driven cavity", "channel flow"};

	sprintf(str, "MAIN_printInfo:\r\n"
	        "\tgrid size      : %i x %i x %i = %.2f * 10^6 Cells\r\n"
	        "\tnTimeSteps     : %i\r\n"
	        "\tresult file    : %s\r\n"
	        "\taction         : %s\r\n"
	        "\tsimulation type: %s\r\n"
	        "\tobstacle file  : %s\r\n\r\n",
	        SIZE_X, SIZE_Y, SIZE_Z, 1e-6*SIZE_X*SIZE_Y*SIZE_Z,
	        param->nTimeSteps, param->resultFilename,
	        actionString[param->action], simTypeString[param->simType],
	        (param->obstacleFilename == NULL) ? "<none>" :
	                                            param->obstacleFilename );
	printf(str);
}

u32 apply_mask(u32 value, u32 mask, u32 data) {
	return (~mask & value) | (mask & data);
}

void disable_core() {
	u32 base = XPAR_PS7_SLCR_0_S_AXI_BASEADDR;
	/* unlock SLCR */
	Xil_Out32(base + 0x8, 0x0000DF0D);

	/* shutdown core #1 */
	u32 a9_cpu = Xil_In32(base + A9_CPU_RST_CTRL);
	Xil_Out32(base + A9_CPU_RST_CTRL, apply_mask(a9_cpu, A9_CLKSTOP1, A9_CLKSTOP1));

	/* lock SLCR */
	Xil_Out32(base + 0x4, 0x0000767B);
}

void slowdown_cpu() {
	/* unlock SLCR */
	Xil_Out32(base + 0x8, 0x0000DF0D);

	/* enable dynamic clock gating */
	u32 pwr_ctrl = 0;
	asm volatile ("mrc p15, 0, %0, c15, c0, 0" : "=r"(pwr_ctrl) : : "cc");
	pwr_ctrl |= 1;
	asm volatile ("mcr p15, 0, %0, c15, c0, 0" : : "r"(pwr_ctrl) : "cc");
	pwr_ctrl = 0;
	asm volatile ("mrc p15, 0, %0, c15, c0, 0" : "=r"(pwr_ctrl) : : "cc");

	/* increase the clock divider */
	u32 arm_clk_ctrl = Xil_In32(base + ARM_CLK_CTRL);
	orig_arm_div = arm_clk_ctrl & DIVISOR;
	arm_clk_ctrl = apply_mask(arm_clk_ctrl, DIVISOR, arm_div);
	Xil_Out32(base + ARM_CLK_CTRL, arm_clk_ctrl);

	u32 fpga0_clk_ctrl = Xil_In32(base + FPGA0_CLK_CTRL);
	orig_fpga0_div0 = fpga0_clk_ctrl & DIVISOR0;
//	orig_fpga0_div1 = fpga0_clk_ctrl & DIVISOR1;
	fpga0_clk_ctrl = apply_mask(fpga0_clk_ctrl, DIVISOR0, fpga0_div0);
//	fpga0_clk_ctrl = apply_mask(fpga0_clk_ctrl, DIVISOR1, fpga0_div1);
	Xil_Out32(base + FPGA0_CLK_CTRL, fpga0_clk_ctrl);

	/* disable unused peripherals */
	Xil_Out32(base + APER_CLK_CTRL, (1<<20) | (1<<21) | (1<<3) | (1<<2));
}

void restore_cpu() {
	/* enable all peripherals */
	Xil_Out32(base + APER_CLK_CTRL, 0xffffffff);

	/* restore fpga0_div */
	u32 fpga0_clk_ctrl = Xil_In32(base + FPGA0_CLK_CTRL);
	fpga0_clk_ctrl = apply_mask(fpga0_clk_ctrl, DIVISOR0, orig_fpga0_div0);
//	fpga0_clk_ctrl = apply_mask(fpga0_clk_ctrl, DIVISOR1, orig_fpga0_div1);
	Xil_Out32(base + FPGA0_CLK_CTRL, fpga0_clk_ctrl);

	/* restore arm_div */
	u32 arm_clk_ctrl = Xil_In32(base + ARM_CLK_CTRL);
	arm_clk_ctrl = apply_mask(arm_clk_ctrl, DIVISOR, orig_arm_div);
	Xil_Out32(base + ARM_CLK_CTRL, arm_clk_ctrl);

	/* disable dynamic clock gating */
	u32 pwr_ctrl = 123;
	asm volatile ("mrc p15, 0, %0, c15, c0, 0" : "=r"(pwr_ctrl) : : "cc");
	pwr_ctrl = apply_mask(pwr_ctrl, 1, 0);
	asm volatile ("mcr p15, 0, %0, c15, c0, 0" : : "r"(pwr_ctrl) : "cc");
	pwr_ctrl = 123;
	asm volatile ("mrc p15, 0, %0, c15, c0, 0" : "=r"(pwr_ctrl) : : "cc");

	/* lock SLCR */
	Xil_Out32(base + 0x4, 0x0000767B);
}
