/* $Id: main.c,v 1.4 2004/04/21 04:23:43 pohlt Exp $ */

/*############################################################################*/

#include "main.h"
#include "lbm.h"
#include <stdio.h>
#include <stdlib.h>

#include <sys/times.h>
#include <unistd.h>

#include <sys/stat.h>
#include <fcntl.h>

/*############################################################################*/
static LBM_GridPtr srcGrid, dstGrid;
static double bus[2*GRID_SIZE];
static double* src_bus;
static double* dst_bus;

/*############################################################################*/

int main( int nArgs, char* arg[] ) {
	MAIN_Param param;
	// redirect stdout to file
	int    fd;
	fpos_t pos;
	int check_results = 0;
	u32 p_srcGrid = 0 + MARGIN * sizeof(double);
	u32 p_dstGrid = GRID_SIZE + MARGIN;
	p_dstGrid *= sizeof(double);

	fflush(stdout);
	fgetpos(stdout, &pos);
	fd = dup(fileno(stdout));
	freopen("lbm.out", "w", stdout);

	nArgs = 6;
	// test workload
	arg[1] = "2";
	arg[2] = "reference.dat";
	arg[3] = "0";
	arg[4] = "0";
	arg[5] = "100_100_130_ldc_tiny.of";

	MAIN_parseCommandLine( nArgs, arg, &param );
	MAIN_printInfo( &param );

	src_bus = bus + p_srcGrid/sizeof(double);
	dst_bus = bus + p_dstGrid/sizeof(double);
	srcGrid = &src_bus;
	dstGrid = &dst_bus;

	MAIN_initialize( &param );

	lbm_acc(bus, &p_srcGrid, &p_dstGrid, &(param.nTimeSteps));
	if (param.nTimeSteps % 2 == 1) {
		LBM_GridPtr aux2 = srcGrid;
		srcGrid = dstGrid;
		dstGrid = aux2;
	}

	MAIN_finalize( &param );

	//recover stdout
	fflush(stdout);
	dup2(fd, fileno(stdout));
	close(fd);
	clearerr(stdout);
	fsetpos(stdout, &pos);        /* for C9X */

	printf("Simulation Complete\n");

	check_results = system("diff size2_nosqrt.out lbm.out");
	if(check_results != 0){
		printf("FAILURE! Output has mismatches with reference output!\n");
		check_results = 1;
	} else{
		printf("SUCCES! Output matches the reference output\n");
	}
	return check_results;
}

/*############################################################################*/

void MAIN_parseCommandLine( int nArgs, char* arg[], MAIN_Param* param ) {
	struct stat fileStat;

	if( nArgs < 5 || nArgs > 6 ) {
		printf( "syntax: lbm <time steps> <result file> <0: nil, 1: cmp, 2: str> <0: ldc, 1: channel flow> [<obstacle file>]\n" );
		exit( 1 );
	}

	param->nTimeSteps     = atoi( arg[1] );
	param->resultFilename = arg[2];
	param->action         = (MAIN_Action) atoi( arg[3] );
	param->simType        = (MAIN_SimType) atoi( arg[4] );

	if( nArgs == 6 ) {
		param->obstacleFilename = arg[5];

		if( stat( param->obstacleFilename, &fileStat ) != 0 ) {
			printf( "MAIN_parseCommandLine: cannot stat obstacle file '%s'\n",
			         param->obstacleFilename );
			exit( 1 );
		}
		if( fileStat.st_size != SIZE_X*SIZE_Y*SIZE_Z+(SIZE_Y+1)*SIZE_Z ) {
			printf( "MAIN_parseCommandLine:\n"
			        "\tsize of file '%s' is %i bytes\n"
					    "\texpected size is %i bytes\n",
			        param->obstacleFilename, (int) fileStat.st_size,
			        SIZE_X*SIZE_Y*SIZE_Z+(SIZE_Y+1)*SIZE_Z );
			exit( 1 );
		}
	}
	else param->obstacleFilename = NULL;
}

/*############################################################################*/

void MAIN_printInfo( const MAIN_Param* param ) {
	const char actionString[3][32] = {"nothing", "compare", "store"};
	const char simTypeString[3][32] = {"lid-driven cavity", "channel flow"};
	printf( "MAIN_printInfo:\n"
	        "\tgrid size      : %i x %i x %i = %.2f * 10^6 Cells\n"
	        "\tnTimeSteps     : %i\n"
	        "\tresult file    : %s\n"
	        "\taction         : %s\n"
	        "\tsimulation type: %s\n"
	        "\tobstacle file  : %s\n\n",
	        SIZE_X, SIZE_Y, SIZE_Z, 1e-6*SIZE_X*SIZE_Y*SIZE_Z,
	        param->nTimeSteps, param->resultFilename,
	        actionString[param->action], simTypeString[param->simType],
	        (param->obstacleFilename == NULL) ? "<none>" :
	                                            param->obstacleFilename );
}

/*############################################################################*/

void MAIN_initialize( const MAIN_Param* param ) {
	LBM_initializeGrid( *srcGrid );
	LBM_initializeGrid( *dstGrid );

	if( param->obstacleFilename != NULL ) {
		LBM_loadObstacleFile( *srcGrid, param->obstacleFilename );
		LBM_loadObstacleFile( *dstGrid, param->obstacleFilename );
	}

	if( param->simType == CHANNEL ) {
		LBM_initializeSpecialCellsForChannel( *srcGrid );
		LBM_initializeSpecialCellsForChannel( *dstGrid );
	} else {
		LBM_initializeSpecialCellsForLDC( *srcGrid );
		LBM_initializeSpecialCellsForLDC( *dstGrid );
	}

	LBM_showGridStatistics( *srcGrid );
}

/*############################################################################*/

void MAIN_finalize( const MAIN_Param* param ) {
	LBM_showGridStatistics( *srcGrid );
}
